<form hx-post="/password" hx-swap="outerHTML" class="row align-items-center" hx-target-error="#toast-content">
    <fieldset>
        <legend>Authentication</legend>
        <div class="col-12 mb-3">
          <label for="password" class="form-label">Password of the user {{ config.username }}</label>
          <input type="password" class="form-control" id="password" name="password" placeholder="Enter password">
        </div>
        <div class="col-12">
            <button type="submit" class="btn btn-primary">Submit</button>
        </div>
    </fieldset>
</form>