"use strict";(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[897],{1889:(t,r,a)=>{a.d(r,{bOW:()=>h});var e=a(3435);function h(t){return(0,e.k5)({tag:"svg",attr:{viewBox:"0 0 24 24",fill:"none",stroke:"currentColor",strokeWidth:"2",strokeLinecap:"round",strokeLinejoin:"round"},child:[{tag:"path",attr:{d:"M14 3v4a1 1 0 0 0 1 1h4"},child:[]},{tag:"path",attr:{d:"M11.5 21h-4.5a2 2 0 0 1 -2 -2v-14a2 2 0 0 1 2 -2h7l5 5v5m-5 6h7m-3 -3l3 3l-3 3"},child:[]}]})(t)}}}]);