<div id="zone">
    <div class="col-12">
        <label for="smt" class="form-label">Password submitted and being checked</label>
    </div>
    <div class="col-12">
        <button id="smt" hx-get="/status" hx-trigger="load delay:5s" hx-target="#zone" class="btn btn-primary">Refresh</button>
    </div>
</div>
