from pyicloud_ipd.services.calendar import CalendarService
from pyicloud_ipd.services.findmyiphone import FindMyiPhoneServiceManager
from pyicloud_ipd.services.ubiquity import UbiquityService
from pyicloud_ipd.services.contacts import ContactsService
from pyicloud_ipd.services.reminders import RemindersService
from pyicloud_ipd.services.photos import PhotosService
from pyicloud_ipd.services.account import AccountService
